local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local plugin = RW:Plugin("Options")

local ACD = LibStub("AceConfigDialog-3.0")
local LoD = RW:GetPlugin("LoD")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("RW-2.0")

local iconCoords = {
	STAR		= {0, 		0.25,	0, 		0.25},
	CIRCLE 		= {0.25,	0.5,	0,		0.25},
	DIAMOND		= {0.5, 	0.75,	0,		0.25},
	TRIANGLE 	= {0.75,	1,		0,		0.25},
	MOON 		= {0,		0.25,	0.25,	0.5},
	SQUARE 		= {0.25,	0.5,	0.25,	0.5},
	CROSS 		= {0.5,		0.75,	0.25,	0.5},
	SKULL 		= {0.75,	1,		0.25,	0.5},
}

local OptionsFrame
local bosses = {}
local bossMods = {}

function plugin:OnRegister()
	
	self.BossOptions = {BossMods = {}}
	self.PluginOptions = {Plugins = {}}
	
	self:RegisterMessage("RW_BossModFound")
	self:RegisterMessage("RW_BossRegistered")
	
	self.options = {
		type = "group",
		name = "Raid Watch 2",
		handler = self,
		args = {
			General = {
				type = "group",
				name = "General",
				order = 1,
				args = {
					Config = {
						type = "execute",
						name = "Config",
						func = "ToggleConfigMode",
					},
				}
			},
			Plugins = {	
				type = "group",
				name = "Plugins",
				order = 3,
				plugins = self.PluginOptions,
				args = {}
			}
		}
	}	
end

function plugin:PluginEnable()
	self.NextConfigMode = true
	LibStub("AceConfig-3.0"):RegisterOptionsTable("RW", self.options)
	
	self:RegisterMessage("RW_BossRegistered")
	self:RegisterMessage("RW_BossModFound")
end

function plugin:PluginDisable()
end

function plugin:ToggleConfigMode()
	RW:SendMessage("RW_ConfigMode", self.NextConfigMode) 
	self.NextConfigMode = not self.NextConfigMode
end

----------------------
-- Message handlers --
----------------------
function plugin:RW_BossRegistered(message, boss)
	bosses[boss.modPack] = bosses[boss.modPack] or {}
	tinsert(bosses[boss.modPack], boss)
end

function plugin:RW_BossModFound(message, modName, modID)
	bossMods[modName] = modID or true
end
----------------------



do
	local dropdown -- Upvalue to access to :DoLayout when needed without having to pass it around everywere as userdata
	local treeStatusTable = {} -- table to store the status of the tree between opening and closing	
	
	local function setOptionValue(widget, event, value)
		local boss, option = widget:GetUserData("boss"), widget:GetUserData("option")
		boss.db.profile[option.optionName] = value
		widget:SetValue(value)		
	end
	
	------------------------------------------------------
	-- Toggle Option
	------------------------------------------------------
	local function getBoolOption(option, boss, parent)
		local check = AceGUI:Create("CheckBox")
		check:SetFullWidth(true)
		check:SetUserData("boss", boss)
		check:SetUserData("option", option)
		check:SetValue(boss.db.profile[option.optionName])
		check:SetCallback("OnValueChanged", setOptionValue)
		check:SetLabel(option.text)
		
		return check
	end
	------------------------------------------------------
	
	------------------------------------------------------
	-- Slider Option
	------------------------------------------------------
	local function getSliderOption(option, boss, parent)
		local check = AceGUI:Create("Slider")
		check:SetFullWidth(true)
		check:SetUserData("boss", boss)
		check:SetUserData("option", option)
		check:SetValue(boss.db.profile[option.optionName])
		check:SetCallback("OnValueChanged", setOptionValue)
		check:SetLabel(option.text)
		
		return check
	end
	------------------------------------------------------
	
	local function getMasterValue(option, boss)
		if boss.db.profile[option.text].uses == 0 then
			return false
		end
		if boss.db.profile[option.text].uses == boss.defaultOptions[option.text].uses then
			return true 
		end
		return nil
	end
	
	-- Enable and Disable the plugin
	local function setPluginEnable(widget, event, value)
		local key, boss, option = widget:GetUserData("key"), widget:GetUserData("boss"), widget:GetUserData("option")
		if bit.band(boss.db.profile[option.text].uses, RW.BITS[key]) == RW.BITS[key] and not value then
			boss.db.profile[option.text].uses = boss.db.profile[option.text].uses - RW.BITS[key]
		else
			boss.db.profile[option.text].uses = boss.db.profile[option.text].uses + RW.BITS[key]
		end
		widget.parent.parent:GetUserData("check"):SetValue(getMasterValue(option, boss))
	end
	
	-- Get the options for each detailed option
	local function getSubOptions(option, boss, oType)
		local enable = AceGUI:Create("CheckBox")
		enable:SetLabel("Enable")
		enable:SetValue(bit.band(boss.db.profile[option.text].uses, RW.BITS[oType]) == RW.BITS[oType])
		enable:SetUserData("key", oType)
		enable:SetUserData("boss", boss)
		enable:SetUserData("option", option)
		enable:SetCallback("OnValueChanged", setPluginEnable)
				
		
		return enable, boss.optionGets[oType](boss, option)	
	end
	
	-- Select one of the tabs in the detailed options
	local function selectOptionDetail(widget, event, group)
		widget:ReleaseChildren()
		local sgroup = AceGUI:Create("ScrollFrame")
		sgroup:SetLayout("Flow")
		sgroup:SetFullWidth(true)
		sgroup:SetFullHeight(true)
		sgroup:SetHeight(150)
		widget:AddChild(sgroup)
		
		sgroup:AddChildren(getSubOptions(widget:GetUserData("option"), widget:GetUserData("boss"), group))
		
		dropdown:DoLayout()
	end
	
	-- Returns the tabs for the details about the selected option
	local tabs = {}
	local function getOptionTabs(option, boss)
		wipe(tabs)
		for i, b in pairs(RW.BITS) do
			if bit.band(option.uses, b) == b then
				tinsert(tabs, {value = i, text = L[i]})
			end
		end
		return tabs
	end
	
	-- Show detailed options for the selection option
	local function showDetails(widget, event)
		local option = widget.parent:GetUserData("option")
		local boss = widget.parent:GetUserData("boss")
		
		local t = getOptionTabs(option, boss)
		
		if #t > 0 then
			local tab = AceGUI:Create("TabGroup")
			tab:SetFullWidth(true)
			tab:SetTabs(t)
			tab:SetUserData("check", widget.parent:GetUserData("check"))
			tab:SetUserData("boss", boss)
			tab:SetUserData("option", option)
			tab:SetCallback("OnGroupSelected", selectOptionDetail)
			tab:SelectTab(t[1] and t[1].value or nil)
			tab:SetCallback("OnTabEnter", function(widget, event, key, tab)
				GameTooltip:SetOwner(widget.frame)
				GameTooltip:AddLine(L[key])
				GameTooltip:AddLine(L[key.."_DESC"])
				GameTooltip:Show()
			end)
			tab:SetCallback("OnTabLeave", function() GameTooltip:Hide() end)
			
			widget.parent:AddChild(tab)
			widget.parent:SetUserData("DetailedOptions", tab)
			widget:SetUserData("detailsGroup", tab)
		else
			
			local heading = AceGUI:Create("InteractiveLabel")
			heading:SetText("No detailed options")
			
			widget.parent:AddChild(heading)
			widget.parent:SetUserData("DetailedOptions", heading)
			widget:SetUserData("detailsGroup", heading)
		end
		
		widget:SetText("Hide details")
		widget:SetCallback("OnClick", widget:GetUserData("hideFunc"))
		dropdown:DoLayout()
	end
	
	-- Hide the detailed options (hack to remove it from the parents children to make it re-layout properly
	local function hideDetails(widget, event)
		local detailsGroup = widget:GetUserData("detailsGroup")
		detailsGroup:Release()
		RW:TRemove(widget.parent.children, detailsGroup)
		widget:SetText("Details")
		widget:SetCallback("OnClick", showDetails)
		dropdown:DoLayout()
	end
	
	-- The checkbox have tri-state. so empty means totally disabled, checked means all settings are default. 
	-- And grey means some settings are not as default but still some enabled
	-- Toggle the option on and off completely
	local function toggleMaster(widget, event, value)
		local boss, option = widget.parent:GetUserData("boss"), widget.parent:GetUserData("option")
		if value == nil then widget:SetValue(false) end
		if value then
			boss.db.profile[option.text].uses = boss.defaultOptions[option.text].uses
		else
			boss.db.profile[option.text].uses = 0
		end
		-- If the detailed options is show then re-create it to reflect changes
		if widget.parent:GetUserData("DetailedOptions") then
			hideDetails(widget.parent:GetUserData("button"), "OnClick")
			dropdown:DoLayout()
			showDetails(widget.parent:GetUserData("button"), "OnClick")
		end
	end
	
	-- Create options for general Option (checkbox, icon, description and show/hide details button)
	local function getGeneralOption(option, boss, parent)
		local group = AceGUI:Create("InlineGroup")
		group:SetFullWidth(true)
		group:SetTitle(option.text)
		group:SetLayout("Flow")
		group:SetUserData("boss", boss)
		group:SetUserData("option", option)
	
		local check = AceGUI:Create("CheckBox")
		check:SetTriState(true)
		check:SetRelativeWidth(0.1)
		check:SetValue(getMasterValue(option, boss))
		check:SetLabel("")
		check:SetCallback("OnValueChanged", toggleMaster)
		
		local button = AceGUI:Create("Button")
		button:SetText("Details")
		button:SetRelativeWidth(0.3)
		button:SetUserData("hideFunc", hideDetails)
		button:SetCallback("OnClick", showDetails)
		
		group:SetUserData("button", button)
		group:SetUserData("check", check)
		group:AddChildren(check, button)
		return group
	end
	
	-- Create options for selected spell (checkbox, icon, description and show/hide details button)
	local function getMasterSpellOption(option, boss, parent)
		local group = AceGUI:Create("InlineGroup")
		group:SetFullWidth(true)
		group:SetTitle(option.text)
		group:SetLayout("Flow")
		group:SetUserData("boss", boss)
		group:SetUserData("option", option)
	
		local check = AceGUI:Create("CheckBox")
		check:SetTriState(true)
		check:SetRelativeWidth(0.1)
		check:SetValue(getMasterValue(option, boss))
		check:SetLabel("")
		check:SetCallback("OnValueChanged", toggleMaster)
		
		local label = AceGUI:Create("InteractiveLabel")
		label:SetRelativeWidth(0.5)
		label:SetImage(option.icon)
		label:SetImageSize(30, 30)
		label:SetCallback("OnEnter", function(self)
			GameTooltip:SetOwner(self.frame)
			GameTooltip:SetHyperlink("spell:"..self.parent:GetUserData("option").spell)
			GameTooltip:Show()
		end)
		label:SetCallback("OnLeave", function()
			GameTooltip:Hide()
		end)
		label:SetCallback("OnClick", function(self, event, button)
			if button == "LeftButton" then
				if IsShiftKeyDown() and ChatFrameEditBox:IsShown() then
					ChatFrameEditBox:Insert("\124cff71d5ff\124Hspell:"..self.parent:GetUserData("option").spell.."\124h["..self.parent:GetUserData("option").text.."]\124h\124r")
				end
			end
		end)
		
		local button = AceGUI:Create("Button")
		button:SetText("Details")
		button:SetRelativeWidth(0.3)
		button:SetUserData("hideFunc", hideDetails)
		button:SetCallback("OnClick", showDetails)
		
		group:SetUserData("button", button)
		group:SetUserData("check", check)
		group:AddChildren(check, label, button)
		return group
	end
	
	-- Select Boss
	local headers = {}
	local function selectBoss(widget, event, group)
		widget:ReleaseChildren()
		boss = bosses[widget:GetUserData("bossMod")][group]
		if not boss then error(("Bossmod %q is not loaded"):format(widget:GetUserData("bossMod"))) end
		
		local scroll = AceGUI:Create("ScrollFrame")
		scroll:SetFullWidth(true)
		scroll:SetFullHeight(true)
		scroll:SetLayout("Flow")
		widget:AddChild(scroll)
		
		-- If this boss uses raid marks, show a box with used icons
		if boss.usedIcons then
			local iconGroup = AceGUI:Create("InlineGroup")
			iconGroup:SetTitle("Used icons")
			iconGroup:SetLayout("Flow")
			scroll:AddChild(iconGroup)
			
			for i in pairs(boss.usedIcons) do
				local icon = AceGUI:Create("Icon")
				icon:SetImage("Interface\\TargetingFrame\\UI-RaidTargetingIcons.blp", unpack(iconCoords[i]))
				icon:SetImageSize(20, 20)
				icon:SetRelativeWidth(0.25)
				icon:SetHeight(25)
				iconGroup:AddChild(icon)
			end
		end
		
		if boss.options then
			-- Build options titles
			wipe(headers)
			if boss.options.headers then
				for header in pairs(boss.options.headers) do
					local group = AceGUI:Create("InlineGroup")
					group:SetLayout("Flow")
					group:SetTitle(header)
					group:SetFullWidth(true)
					scroll:AddChild(group)
					headers[header] = group
				end
			end
			-- Add all the options
			for optionName, option in pairs(boss.options) do
				if option.type == "spell" then
					if option.header and headers[option.header] then
						headers[option.header]:AddChildren(getMasterSpellOption(option, boss, widget))
					else
						scroll:AddChildren(getMasterSpellOption(option, boss, widget))
					end			
				elseif option.type == "bool" then
					if option.header and headers[option.header] then
						headers[option.header]:AddChildren(getBoolOption(option, boss, widget))
					else
						scroll:AddChildren(getBoolOption(option, boss, widget))
					end
				elseif option.type == "slider" then
					if option.header and headers[option.header] then
						headers[option.header]:AddChildren(getSliderOption(option, boss, widget))
					else
						scroll:AddChildren(getSliderOption(option, boss, widget))
					end
				elseif option.type == "general" then
					if option.header and headers[option.header] then
						headers[option.header]:AddChildren(getGeneralOption(option, boss, widget))
					else
						scroll:AddChildren(getGeneralOption(option, boss, widget))
					end
				end
			end
		end
	end
	
	bossList = {}
	local function buildBossList(modName, subGroup)
		local firstNr = nil
		wipe(bossList)
		for	i, boss in pairs(bosses[modName]) do
			if boss.subGroup == subGroup then
				if not firstNr then firstNr = i end
				bossList[i] = boss:GetName()
			end
		end
		return bossList, firstNr
	end
	
	local function loadFunc(widget, event)
		if not IsAddOnLoaded(widget:GetUserData("modID")) then
			LoD:LoadMod(widget:GetUserData("modID"), widget:GetUserData("modName"))
			widget:GetUserData("tree"):SetTree(plugin:BuildOptionsTree())
			selectOption(widget.parent, "OnGroupSelected", widget:GetUserData("path"))
		end
	end
	

	-- Select Option
	local nodes
	local function selectOption(widget, event, group)
		widget:ReleaseChildren()		
		
		-- Only make AceGUI configuration if its a boss mod
		-- Else open host the AceConfig inside this window
		if select(1, string.split("\001", group)) ~= "BossMods" then
			local bgroup = AceGUI:Create("SimpleGroup")
			bgroup:SetLayout("Fill")
			bgroup:SetFullWidth(true)
			bgroup:SetFullHeight(true)
			widget:AddChild(bgroup)
			ACD:Open("RW", bgroup, string.split("\001", group))
			
		else
			-- Get the full path minus "BossMods"
			nodes = { select(2, string.split("\001", group)) }
			
			local bossMod, subGroup
			-- This mod has sub groups
			if #nodes > 1 then
				bossMod = select(4, nodes[#nodes - 1]:find("type:([%w%s]*);?([%w%s]*)"))
				subGroup = select(4, nodes[#nodes]:find("type:([%w%s]*);?([%w%s]*)"))
			-- This mod have no subgroups
			else
				bossMod = select(4, nodes[#nodes]:find("type:([%w%s]*);?([%w%s]*)"))
			end
			-- This mod is loaded
			if bosses[bossMod] then
				nameList, firstNr = buildBossList(bossMod, subGroup)
				if firstNr then
					dropdown = AceGUI:Create("DropdownGroup")
					dropdown:SetLayout("Flow")
					dropdown:SetUserData("bossMod", bossMod)
					dropdown:SetFullWidth(true)
					dropdown:SetFullHeight(true)
					dropdown:SetGroupList(nameList)
					dropdown:SetCallback("OnGroupSelected", selectBoss)
					dropdown:SetGroup(firstNr)
					widget:AddChild(dropdown)
				end
			-- Not loaded, (add load button)
			else
				local button = AceGUI:Create("Button")
				button:SetText("Load")
				button:SetUserData("modID", bossMods[bossMod])
				button:SetUserData("modName", bossMod)
				button:SetUserData("path", group)
				button:SetUserData("tree", widget)
				button:SetCallback("OnClick", loadFunc)
				widget:AddChild(button)
			end
		end
	end
	
	-- Toggle the main options window
	function plugin:Toggle()
		if OptionsFrame then
			OptionsFrame:ReleaseChildren()
			OptionsFrame:Release()
			OptionsFrame = nil
		else		
			OptionsFrame = AceGUI:Create("Frame")
			OptionsFrame:SetTitle("Raid Watch 2")
			OptionsFrame:SetLayout("Fill")
			OptionsFrame:SetWidth(1000)
			OptionsFrame:SetHeight(800)
			OptionsFrame:SetCallback("OnClose", function(widget)
				widget:ReleaseChildren()
				widget:Release()
				OptionsFrame = nil
			end)
			
			local tree = AceGUI:Create("TreeGroup")
			tree:SetLayout("Flow")
			tree:SetTree(self:BuildOptionsTree())
			tree:SetCallback("OnGroupSelected", selectOption)			
			tree:SetStatusTable(treeStatusTable)
			OptionsFrame:AddChild(tree)
			
			OptionsFrame:Show()
		end
	end
end

do
	local TreeGroups = {}
	function plugin:BuildOptionsTree()
		wipe(TreeGroups)
		TreeGroups = {
			{
				value = "General",
				text = "General"
			},
			{
				value = "BossMods",
				text = "Boss Mods",
				children = {
				}
			},
			{
				value = "Plugins",
				text = "Plugins",
				children = {
				}
			}
		}
		for i, module in RW:IteratePlugins() do
			if module.GetOptions then
				self.PluginOptions.Plugins[module:GetName()] = module:GetOptions()
				tinsert(TreeGroups[3].children, {value = module:GetName(), text = module:GetName()})
			end
		end
		
		local function checkList(lst, value)
			for i, t in pairs(lst) do
				if t.text == value then return false end
			end
			return true
		end
		
		for modName, modID in pairs(bossMods) do
			leafe = {value = "type:leafe;"..modName, text = modName}
			if bosses[modName] then
				for i, boss in pairs(bosses[modName]) do
					if boss.subGroup then
						if not leafe.children or checkList(leafe.children, boss.subGroup) then
							leafe.value = "type:subGroup;"..modName
							leafe.children = leafe.children or {}
							tinsert(leafe.children, {value = "type:leafe;"..boss.subGroup, text = boss.subGroup})
						end
					end
				end
			end		
			tinsert(TreeGroups[2].children, leafe)
		end
				
		return TreeGroups
	end
end
