local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local plugin = RW:Plugin("MessageFrame", "LibSink-2.0")
if not plugin then return end

local L = LibStub("AceLocale-3.0"):GetLocale("RW-2.0-Plugins")

local LSM = LibStub("LibSharedMedia-3.0")

local nrSlots = 5

function plugin:OnRegister()
	local defaults = {
		profile = {
			GrowUp = true,
			Font = "MyriadSBC",
			FontSize = 20,
			FadeOutTime = 10,
			SlotHeight = 30,
			sink20OutputSink = "RW",
			ScaleUpTime = 0.3,
			ScaleDownTime = 0.6
		}
	}
	self.db = RW.db:RegisterNamespace(self:GetName(), defaults)
	self:SetSinkStorage(self.db.profile)
	self:RegisterSink("RW", "Raid Watch", "Messages trough RW", "AddMessage")
end

function plugin:PluginEnable()
	self:RegisterMessage("RW_ConfigMode")
	self:RegisterMessage("RW_Message")
	self:RegisterMessage("RW_DefaultsChanged")
end

function plugin:PluginDisable()
	self.slots = nil
end


local function dragStart(self)
	self:StartMoving()
end
local function dragStop(self)
	self:StopMovingOrSizing()
	local s = self:GetEffectiveScale()
	plugin.db.profile.x = self:GetLeft() * s
	plugin.db.profile.y = self:GetBottom() * s
end

local function onUpdate(self, elapsed)
	local db = plugin.db.profile
	local showing = false
	for i, s in next, plugin.slots do
		if s:IsShown() then
			showing = true
			s.ScaleTime = s.ScaleTime + elapsed
			if s.ScaleTime < db.ScaleUpTime then
				s:SetScale(s:GetScale() + elapsed)
				s.text:SetTextHeight(s.text:GetStringHeight() + 50 * elapsed)
			elseif s.ScaleTime <= db.ScaleDownTime then
				s:SetScale(s:GetScale() - elapsed)
				s.text:SetTextHeight(s.text:GetStringHeight() - 50 * elapsed)
			else
				s:SetScale(1)
				s.text:SetFont(LSM:Fetch("font", db.Font), db.FontSize, "OUTLINE")
			end
			FadingFrame_OnUpdate(s)
		else
		end
	end
	if not showing then self:SetScript("OnUpdate", nil) end
end
	
function plugin:CreateAnchor()		
	self.anchor = RW:CreateAnchor("MessageAnchor", UIParent, "Messages")
	local s = self.anchor:GetEffectiveScale()
	self.anchor:SetScript("OnDragStart", dragStart)
	self.anchor:SetScript("OnDragStop", dragStop)
	if self.db.profile.x then
		self.anchor:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", self.db.profile.x / s, self.db.profile.y / s)
	else
		self.anchor:SetPoint("TOP", RaidWarningFrame, "BOTTOM", 0, -50)
	end
	
	local frame = CreateFrame("Frame", nil, UIParent)	
	self.eventFrame = frame
end

function plugin:Reset()
	self.db.profile.x = nil
	self.db.profile.y = nil
	if self.anchor then 
		self.anchor:ClearAllPoints()
		self.anchor:SetPoint("TOP", RaidWarningFrame, "BOTTOM", 0, -50) 
	end
end

function plugin:GetMessageFrame()
	if not self.anchor then self:CreateAnchor() end
	local db = self.db.profile
	
	if not self.slots then
		self.slots = {}
		for i = 1, nrSlots do
			local frame = CreateFrame("Frame", "SlotFrame_"..i, self.eventFrame)
			frame:SetWidth(1)
			frame:SetHeight(db.SlotHeight)
			local str = frame:CreateFontString("RWMessageSlot_"..i, "ARTWORK")
			str:SetPoint("CENTER")
			str:SetFont(LSM:Fetch("font", db.Font), db.FontSize, "OUTLINE")
			
			local icon = frame:CreateTexture()
			icon:SetPoint("RIGHT", str, "LEFT", -4, 0)
			icon:SetWidth(db.SlotHeight - 4)
			icon:SetHeight(db.SlotHeight - 4)
			
			FadingFrame_SetFadeInTime(frame, 0.4)
			FadingFrame_SetFadeOutTime(frame, 3)
			
			frame.text = str
			frame.icon = icon
			
			frame:Hide()
			self.slots[i] = frame
		end
	end

	local slots = self.slots
	local old = slots[#slots]
	for i = #slots, 1, -1 do
		if i > 1 then
			slots[i] = slots[i-1]
		else
			slots[i] = old
		end
	end
	
	for i, t in pairs(self.slots) do
		t:ClearAllPoints()
		if i == 1 then
			t:SetPoint(db.GrowUp and "BOTTOM" or "TOP", self.anchor, db.GrowUp and "TOP" or "BOTTOM", 0, db.GrowUp and 2 or -2)
		else
			t:SetPoint(db.GrowUp and "BOTTOM" or "TOP", self.slots[i-1], db.GrowUp and "TOP" or "BOTTOM", 0, db.GrowUp and 2 or -2)
		end
		FadingFrame_SetHoldTime(t, db.FadeOutTime)
	end	

	return self.slots[1]
end

function plugin:AddMessage(addon, text, r, g, b, font, size, outline, sticky, loc, icon)
	local slot = self:GetMessageFrame()
	slot.text:SetText(text)
	slot.text:SetTextColor(r, g, b)
	slot.icon:SetTexture(icon)
	slot.ScaleTime = 0
	FadingFrame_Show(slot)
	if not self.eventFrame:GetScript("OnUpdate") then self.eventFrame:SetScript("OnUpdate", onUpdate) end
end

------------------------
-- Message handling ----
------------------------
function plugin:RW_Message(message, text, r, g, b, a, icon)
	self:Pour(text or "Test message", r, g, b, nil, nil, nil, nil, nil, icon)
end

function plugin:RW_ConfigMode(message, start)
	local function randomMessage()
		plugin:RW_Message("RW_Message", L["RANDOM_MESSAGE_"..math.random(1, 4)], math.random(), math.random(), math.random(), math.random())
	end
	if not self.anchor then self:CreateAnchor() end
	if start then
		self.anchor:Show()
		self:ScheduleRepeatingTimer(randomMessage, 2)
	else
		self.anchor:Hide()
		self:CancelAllTimers()
	end
end

function plugin:RW_DefaultsChanged(message, option, value)
	if option == "Font" then
		self.db.profile.Font = value
	end
	if self.slots then
		for i, slot in pairs(self.slots) do
			slot.text:SetFont(LSM:Fetch("font", self.db.profile.Font), self.db.profile.FontSize, "OUTLINE")
		end
	end
end
---------------------------------------------

function plugin:GetOptions()
	local db = self.db.profile
	local options = {
		name = "Message Frame",
		type = "group",
		set = function(info, value)
			db[info[#info]] = value
		end,
		get = function(info)
			return db[info[#info]]
		end,
		args = {
			TargetFrame = plugin:GetSinkAce3OptionsDataTable(),
			FadeOutTime = {
				type = "range",
				name = L.FADE_OUT_TIME,
				desc = L.FADE_OUT_TIME_DESC,
				step = 1,
				max = 15,
				min = 2,
			},
			GrowUp = {
				type = "toggle",
				name = L.GROW_UP,
				desc = L.GROW_UP_DESC,
			},
			Reset = {
				type = "execute",
				name = L.RESET_POSITION,
				desc = L.RESET_POSITION_DESC,
				func = function() plugin:Reset() end,
			},
		}
	}
	
	options.args.TargetFrame.guiInline = true
	
	return options
end

